# fsco.li Status

Maintainer: Jakob Moser <moser@cl.uni-heidelberg.de>

[🔗 Website](https://status.fsco.li) that shows the status of several web services hosted by the Fachschaft Computerlinguistik. Consists of a JavaScript-powered HTML frontend and a PHP-powered backend which communicate using an API.

## Run locally

To test the application locally, you need to start a webserver (e.g. Apache) that serves content from this repository and supports PHP.

Using Docker, you can do the following:

```
git clone https://gitlab.cl.uni-heidelberg.de/fachschaft/status.git
cd status
sudo docker compose up
```

The site is then available at http://localhost:8080

## Architecture

The application consists of two parts: A client-side part, written in HTML, CSS and Javascript, and a server-side part, written in PHP. They communicate using an API.

### The client side

```
.
├── 📄 index.html
├── 📁 css
│   └── 🎨 style.css
├── 📁 img
│   ├── 🖼️ dropdown.png
│   ├── 🖼️ favicon.png
│   ├── 🖼️ header.jpg
│   ├── 🖼️ logo.png
│   ├── 🖼️ news.png
│   ├── 🖼️ rocket.png
│   └── 🖼️ ticker.png
└── 📁 js
    ├── 🔑 .htaccess
    └── 🧩 index.mjs
```

The client-side code consists mainly of `index.html` and `index.mjs`, accompanied by some CSS and images for styling. 

All these files are _static_, meaning that they are not in any way interpreted by the server, but just sent to the client (i.e. your browser) as is. This is generally nice for scenarios where load balancing is important, because static files can be cached and easily distributed, after all, there is no central database that could be a bottleneck.

Static sites can also be nicely (and freely) hosted using [GitLab Pages](https://docs.gitlab.com/ee/user/project/pages/) or [GitHub Pages](https://pages.github.com/). Several people use static sites for blogging. “But wait, isn't a blog dynamic? After all, new blog posts are added all the time.”, you might wonder, and you wouldn't really be mistkane: A blog _feels_ very dynamic, but actually only needs to be changed whenever you want to add a blog post. For the rest of the time, it is static. As long as you are comfortable with editing an HTML file (or some other kind of text file) when posting, a blog is actually a pretty static thing.

Back to the status page. “Hey now, a status page is definitely dynamic. How should this work with a static website?”, you might wonder, and you wouldn't be mistaken: With an entirely static website, it doesn't (unless you are comfortable with editing the status page manually every time there is an update).

This is why the client side here actually does not contain any data. The JavaScript program `index.mjs` makes HTTPS requests to the server side (which is very dynamic) over an API, which returns the data the program than integrates into the website for displaying in your browser. The `.htaccess` file is just there so that the Apache webserver sends the right content type header (`text/javascript`) when sending an `*.mjs` file, otherwise the browser would refuse to interpret it as JavaScript.

### The API

So how do those requests look like? And how do the replies from the server look like?

A request looks like this:

`GET` https://status.fsco.li/api/v1/services

And a response like this:

```json
[
  {
    "name": "Website",
    "host": "fachschaft.cl.uni-heidelberg.de",
    "status": 200,
    "category": "public"
  },
  {
    "name": "Tickets",
    "host": "tickets.fachschaft.cl.uni-heidelberg.de",
    "status": 404,
    "category": "public"
  }
]
```

The response is a list of objects in the JSON format (very similary to a list of dictionaries in Python). For every service, it lists the name, the host, the status and a category that the JavaScript script will use to build the webpage that is ultimately displayed by your browser.

You might wonder if there is a more standardized way to specify how an API looks like than giving some examples in a continuous text. And you would be right, there is a way:

```
.
├── 📁 api
│   └── 📁 v1
│       ├── 📄 index.html
│       └── 📋 openapi-spec.yaml
└── 📁 lib
    └── 📁 redoc
        ├── 📄 LICENSE
        ├── 🧩 redoc.standalone.js
        └── 📄 redoc.standalone.js.LICENSE.txt
```

The file `api/v1/openapi-spec.yaml` contains a description of the API in a machine-readable format (namely the OpenAPI format). This can be rendered by various tools, e.g., the Swagger Editor or Redoc, or the GitLab-integrated viewer.

Of course, displaying them is not the only thing you can do with OpenAPI files. You could also validate check for consistency, or even automatically generate clients and servers for a given API as described in such a file.

The `api/v1/index.html` and everything in `lib/redoc` is static code that renders the OpenAPI specification for a visitor. This means you can comfortably look at the API at https://status.fsco.li/api/v1.

* [Swagger Editor](https://editor.swagger.io/)
* [Redoc](https://redocly.github.io/redoc/)

### The server side

```
.
└── 📁 api
    ├── 🔑 .htaccess
    └── 📁 v1
        ├── 🔑 .htaccess
        └── 🐘 services.php
```

The main part of the server-side code is `api/v1/services.php`. Whenever a client makes a request to the server at this path, the server executes the PHP code. The PHP code can then output things (using `echo`) which are sent back to the client. Unlike static files (that are sent to the client in verbatim), the actual code contained in the PHP file is never sent back to the client[^1].

The code itself sends HTTPS requests to the different services to see if they are up and crafts the JSON response as described in the OpenAPI specification.

The `.htaccess` files are configuration for the Apache webserver:

* `api/.htaccess` redirects every request to `/api` to `/api/v1` (so you can enter https://status.fsco.li/api in your browser and are automatically directed to https://status.fsco.li/api/v1). This is just a comfort feature.
* `api/v1/.htaccess` rewrites every request to `/api/v1/services` to go to `/api/v1/services.php`.
  * This is necessary because we want to provide a true REST-API experience (where all the request URLs look like paths), so we would like to avoid the `.php` suffix.

### Docker

```
.
├── 🐋 docker-compose.yml
└── 🐋 Dockerfile
```

Only for development purposes, this repository contains a Docker configuration, so you can quickly start an Apache webserver locally to preview the site.

Why only for development? If you have a look at the `Dockerfile`, you see that it actually doesn't `COPY` any code. It just sets an Apache configuration option and then calls it a day. The `docker-compose.yml` is where we dynamically mount the contents of this repository into the container. This is very common for development: Any change you make to the repository is immediately reflected inside the container, so you don't need to rebuild the image every time you make a change.

While very common for development, it is very uncommon for production to only mount the code: The idea of an image is precisely to be self-contained, so it should contain all necessary code by itself (also allowing you to quickly switch between versions of the code by switching between images). To build a production-ready setup, you should add a `COPY . /var/www/html` instruction to the `Dockerfile` and remove the `volumes` section from `docker-compose.yml`.

## Deployment

For deployment to production, we actually use a more “classic” way (not Docker), namely webspace provided by the hoster lima-city, accessible via FTP. This means we get username and password from our hoster, log in using an FTP client and copy all the files over.

Why do we do it this way and not using Docker? Because the “classic” is what lima-city offers for free: They probably host one large Apache webserver, use the Apache feature of “virtual hosts“ to create a subdirectory for every user and hand out access to this. That way, they don't have to bother with Docker.

Why didn't we host the status page on our own infrastructure (and then using Docker)? Because it would be nice if the status page were still available even when the Fachschaft server is down.

* [lima-city](https://www.lima-city.de/?cref=435320) (referral link[^2])

## Questions & Answers

### The PHP code doesn't look very idiomatic, does it?

You're probably right. PHP is often mixed together with HTML to form some kind of templating system where the `.php` file directly produces HTML that is sent to and displayed in the browser.

This is also how PHP is explained in many tutorials, e.g. this one:

* [PHP: Your first PHP-enabled page - Manual](https://www.php.net/manual/en/tutorial.firstpage.php)

However, nothing about PHP (the language) mandates that it has to output HTML. If you prefer to have a clear separation of concerns (server code only supplies data in machine-readable format, client code deals with presenting that to the user), you can write a “modern-style” API returning JavaScript in any language, including PHP.

One goal of this project is to show that this is possible.

### Why the convoluted directory structure for the API?

Why is the PHP code located in `/api/v1/services.php` and not simply in a file `/services.php`?

First, it is common to separate the API routes of a web application from the routes that e.g. serve the frontend, therefore `/api/`.

Second, you should version your API, e.g. using Semantic Versioning (version numbers in the format `MAJOR.MINOR.PATCH`, e.g. `1.0.0`). It is then common to add the major version number to the API path. If you make incompatible changes to your API, you increase the major version number. This allows you to easily keep the old version of the API as long as some clients still need it. Therefore `v1/`.

This is why we have chosen a more convoluted-looking directory structure instead of just placing the file at root level.

* [Semantic Versioning](https://semver.org/)

### Couldn't we have done this entirely on the client?

You might have noticed that we just asserted we had to make the requests to the Fachschaft services using server side code. However, isn't it possible to send requests in plain, client-side JavaScript? After all, we need to send a request to the API. So couldn't we have implemented the status website as entirely static, with only client-side code?

Yes, almost. 

The problem is that due to the Same-Origin-Policy, client-side JavaScript cannot access responses to requests it sends to different origins (i.e. websites). So the browser will not permit JavaScript running under https://status.fsco.li to get responses from e.g. https://fachschaft.cl.uni-heidelberg.de.

Why does this policy exist? Good question. The main reason is to protect against a security hole that is only there for historical reasons (but it is there). The policy could have been avoided if the web was designed differently, but sadly it wasn't, so we have to deal with that.

* [The Same Origin Policy - Hacker History (video)](https://www.youtube.com/watch?v=bSJm8-zJTzQ)

## License

This project is released into the [Public Domain (CC0)](LICENSE). You can do with it whatever you want and you don't need to credit anyone.

Note that this does not apply to third party files (those found in `lib/*`), namely:

* Redoc (`lib/redoc/*`), which is licensed under the [MIT License](lib/redoc/LICENSE)
* fscoli-next Theme (`lib/fscoli-next/*`), which is licensed under the [GNU General Public License v3.0](lib/fscoli-next/style.css)

For details and authorship information, see the linked license files.

## Footnotes

[^1]: That is, unless you make a configuration error or run a server without PHP capabilities. This means: Yes, you can technically store secrets in PHP files, they should not be sent to the client – but if you ever do that, make sure you verify everything works as it should.
[^2]: Jakob gets commission if you happen to pay for a service at lima-city via this link. Not paid, Jakob just actually likes lima-city.
